#include "stm32f0xx_conf.h"

void USART2_SendString(const char *str);
void USART2_IRQHandler(void);
void USART2_SendHexByte(uint8_t c);
void HandleMacCommand();
void SendMacResponse(uint8_t respbyte);
uint16_t PollPS2Frame();
void SendPS2Command(uint8_t cmd);
void SetPS2LEDState(uint8_t cmd);
uint8_t PS2ToMacScancode(uint8_t code, uint8_t keyup, uint8_t extended);

static __IO unsigned long tenustimer;
static __IO uint8_t mackey;

#define kPS2KbdDataPin              GPIO_Pin_7
#define kPS2KbdClockPin             GPIO_Pin_6
#define kMacKbdDataPin              GPIO_Pin_9
#define kMacKbdClockPin             GPIO_Pin_8

#define kMacKbdStateIdle            0
#define kMacKbdStateReadSettle      1
#define kMacKbdStateReading         2
#define kMacKbdStateWriteSettle     3
#define kMacKbdStateWriting         4
#define kMacKbdStateInquiryWait     5

struct mackbd {
	uint8_t state;
	uint8_t bits;
	uint8_t clockstate;
	uint32_t delay;
	uint8_t cmd;
	uint8_t response;
};
static __IO struct mackbd mackbd;

void SysTick_Handler(void) {
	GPIO_InitTypeDef GPIO_InitStructure;
	uint8_t respbit;
	tenustimer++;

	switch(mackbd.state) {
	case kMacKbdStateIdle:
		if(GPIO_ReadInputDataBit(GPIOC, kMacKbdDataPin) == 0) {
			mackbd.state = kMacKbdStateReadSettle;
			mackbd.delay = 18;
			mackbd.bits = 0;
			mackbd.cmd = 0;
			mackbd.clockstate = 3;
		}
		break;
	case kMacKbdStateReadSettle:
		if(--mackbd.delay == 0) {
			mackbd.state = kMacKbdStateReading;
		}
		break;
	case kMacKbdStateReading:
		switch(mackbd.clockstate) {
		case 0:
			if(--mackbd.delay == 0) {
				mackbd.cmd |= GPIO_ReadInputDataBit(GPIOC, kMacKbdDataPin) << (7 - mackbd.bits++);
				mackbd.delay = 22;
				mackbd.clockstate = 1;
				GPIO_SetBits(GPIOC, kMacKbdClockPin);
			}
			break;
		case 1:
			if(--mackbd.delay == 0) {
				if(mackbd.bits == 8) {
					// Done clocking in all the bits
					mackbd.state = kMacKbdStateIdle;
					HandleMacCommand();
				} else {
					// Not done, transition to next clock
					mackbd.delay = 18;
					mackbd.clockstate = 0;
					GPIO_ResetBits(GPIOC, kMacKbdClockPin);
				}
			}
			break;
		case 3:
			// We just entered the read state, bring the clock low
			mackbd.clockstate = 0;
			mackbd.delay = 18;
			GPIO_ResetBits(GPIOC, kMacKbdClockPin);
			break;
		default:
			break;
		};
		break;
	case kMacKbdStateWriteSettle:
		if(--mackbd.delay == 0) {
			mackbd.state = kMacKbdStateWriting;
			mackbd.clockstate = 3;
			mackbd.bits = 0;
		}
		break;
	case kMacKbdStateWriting:
		switch(mackbd.clockstate) {
		case 0:
			if(--mackbd.delay == 0) {
				mackbd.delay = 18;
				mackbd.clockstate = 1;
				GPIO_SetBits(GPIOC, kMacKbdClockPin);
			}
			break;
		case 1:
			if(--mackbd.delay == 0) {
				if(mackbd.bits == 8) {
					mackbd.state = kMacKbdStateIdle;
					GPIO_InitStructure.GPIO_Pin = kMacKbdDataPin;
					GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
					GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
					GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
					GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
					GPIO_Init(GPIOC, &GPIO_InitStructure);
				} else {
					respbit = (mackbd.response >> (7-mackbd.bits)) & 0x01;
					mackbd.delay = 16;
					mackbd.clockstate = 0;
					GPIO_ResetBits(GPIOC, kMacKbdClockPin);
					respbit ? GPIO_SetBits(GPIOC, kMacKbdDataPin) : GPIO_ResetBits(GPIOC, kMacKbdDataPin);
					mackbd.bits++;
				}
			}
			break;
		case 3:
			GPIO_InitStructure.GPIO_Pin = kMacKbdDataPin;
			GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
			GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
			GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
			GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
			GPIO_Init(GPIOC, &GPIO_InitStructure);

			respbit = (mackbd.response >> (7-mackbd.bits)) & 0x01;
			mackbd.bits++;
			mackbd.clockstate = 0;
			mackbd.delay = 16;
			GPIO_ResetBits(GPIOC, kMacKbdClockPin);
			respbit ? GPIO_SetBits(GPIOC, kMacKbdDataPin) : GPIO_ResetBits(GPIOC, kMacKbdDataPin);
			break;
			
		default:
			break;
		};
		break;
	case kMacKbdStateInquiryWait:
		if(mackey) {
			mackbd.state = kMacKbdStateWriteSettle;
			mackbd.delay = 1;
			mackbd.response = mackey;
			mackey = 0;
		} else if(--mackbd.delay == 0) {
			// If no keyboard press, send a null
			mackbd.state = kMacKbdStateWriteSettle;
			mackbd.delay = 1;
			mackbd.response = 0x7B;
		}
		break;
	default:
		break;
	};
}

int main(void)
{
	GPIO_InitTypeDef GPIO_InitStructure;
	USART_InitTypeDef USART_InitStructure;
	NVIC_InitTypeDef NVIC_InitStructure;
	uint8_t ps2ledstate = 0;

	tenustimer = 0;
	mackey = 0;

	RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOC, ENABLE);

	// Mac Keyboard
	// Clock
	GPIO_InitStructure.GPIO_Pin = kMacKbdClockPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_NOPULL;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	// Data
	GPIO_InitStructure.GPIO_Pin = kMacKbdDataPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	// PS2 keyboard
	GPIO_InitStructure.GPIO_Pin = kPS2KbdDataPin | kPS2KbdClockPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	// UART2
	RCC_AHBPeriphClockCmd(RCC_AHBPeriph_GPIOA, ENABLE);
	RCC_APB1PeriphClockCmd(RCC_APB1Periph_USART2,ENABLE);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource2, GPIO_AF_1);
	GPIO_PinAFConfig(GPIOA, GPIO_PinSource3, GPIO_AF_1);
	GPIO_InitStructure.GPIO_Pin =  GPIO_Pin_2 | GPIO_Pin_3;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_AF;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOA, &GPIO_InitStructure);

	NVIC_InitStructure.NVIC_IRQChannel = USART2_IRQn;
	NVIC_InitStructure.NVIC_IRQChannelPriority = 0;
	NVIC_InitStructure.NVIC_IRQChannelCmd = ENABLE;
	NVIC_Init(&NVIC_InitStructure);

	USART_InitStructure.USART_BaudRate = 9600;
	USART_InitStructure.USART_WordLength = USART_WordLength_8b;
	USART_InitStructure.USART_StopBits = USART_StopBits_1;
	USART_InitStructure.USART_Parity = USART_Parity_No;
	USART_InitStructure.USART_HardwareFlowControl = USART_HardwareFlowControl_None;
	USART_InitStructure.USART_Mode = USART_Mode_Rx | USART_Mode_Tx;
	USART_Init(USART2, &USART_InitStructure);
	USART_Cmd(USART2,ENABLE);
	USART_ITConfig(USART2, USART_IT_RXNE, ENABLE);

	USART2_SendString("Starting up\n\r");
	GPIO_SetBits(GPIOC, kMacKbdClockPin);

	mackbd.state = kMacKbdStateIdle;

	// set a ~10us timer
	SysTick_Config(SystemCoreClock/100000);
	USART2_SendString("Timer configured\n\r");

	while(1) {
		uint16_t ps2frame = 0;
		uint8_t ps2scancode;

		ps2frame = PollPS2Frame();
		ps2scancode = ps2frame >> 1;
		//USART2_SendString("\n\r");

		if(ps2scancode == 0xF0) {
			// Keyup, poll in the scancode of the key that has been released
			ps2frame = PollPS2Frame();
			ps2scancode = ps2frame >> 1;
			mackey = PS2ToMacScancode(ps2scancode, 1, 0);
		} else if(ps2scancode == 0xE0) {
			// two byte scan code
			ps2frame = PollPS2Frame();
			ps2scancode = ps2frame >> 1;
			if(ps2scancode == 0xF0) {
				ps2frame = PollPS2Frame();
				ps2scancode = ps2frame >> 1;
				if(ps2scancode == 0x7C) {
					// prntscrn
					ps2frame = PollPS2Frame();
					ps2scancode = ps2frame >> 1;
					// should be E0
					ps2frame = PollPS2Frame();
					ps2scancode = ps2frame >> 1;
					// should be F0
					ps2frame = PollPS2Frame();
					ps2scancode = ps2frame >> 1;
					// should be 12
				} else {
					mackey = PS2ToMacScancode(ps2scancode, 0, 1);
				}
			} else if(ps2scancode == 0x12) {
				// prntscrn
				ps2frame = PollPS2Frame();
				ps2scancode = ps2frame >> 1;
				// should be E0
				ps2frame = PollPS2Frame();
				ps2scancode = ps2frame >> 1;
				// should be 7C
			} else {
				mackey = PS2ToMacScancode(ps2scancode, 0, 1);
			}
		} else {
			// If it's capslock, turn on the capslock led
			if((ps2scancode) == 0x58) {
				if(ps2ledstate & (1<<2)) {
					ps2ledstate &= ~(1<<2);
				} else {
					ps2ledstate |= 1<<2;
				}
				SetPS2LEDState(ps2ledstate);
			}
			mackey = PS2ToMacScancode(ps2scancode, 0, 0);
		}
		USART2_SendHexByte(ps2scancode);
		
	}

}

uint8_t PS2ToMacScancode(uint8_t code, uint8_t keyup, uint8_t extended)
{
	uint8_t maccode = 0;
	if(!extended) {
		switch(code) {
		case 0x1C: maccode = 0x01; break; // A
		case 0x32: maccode = 0x17; break; // B
		case 0x21: maccode = 0x11; break; // C
		case 0x23: maccode = 0x05; break; // D
		case 0x24: maccode = 0x1D; break; // E
		case 0x2B: maccode = 0x07; break; // F
		case 0x34: maccode = 0x0B; break; // G
		case 0x33: maccode = 0x09; break; // H
		case 0x43: maccode = 0x45; break; // I
		case 0x3B: maccode = 0x4D; break; // J
		case 0x42: maccode = 0x51; break; // K
		case 0x4B: maccode = 0x4B; break; // L  // woah
		case 0x3A: maccode = 0x5D; break; // M
		case 0x31: maccode = 0x5B; break; // N
		case 0x44: maccode = 0x3F; break; // O
		case 0x4D: maccode = 0x47; break; // P
		case 0x15: maccode = 0x19; break; // Q
		case 0x2D: maccode = 0x1F; break; // R
		case 0x1B: maccode = 0x03; break; // S
		case 0x2C: maccode = 0x23; break; // T
		case 0x3C: maccode = 0x41; break; // U
		case 0x2A: maccode = 0x13; break; // V
		case 0x1D: maccode = 0x1B; break; // W
		case 0x22: maccode = 0x0F; break; // X
		case 0x35: maccode = 0x13; break; // Y
		case 0x1A: maccode = 0x0D; break; // Z
		case 0x45: maccode = 0x3B; break; // 0
		case 0x16: maccode = 0x25; break; // 1
		case 0x1E: maccode = 0x27; break; // 2
		case 0x26: maccode = 0x29; break; // 3
		case 0x25: maccode = 0x2B; break; // 4
		case 0x2E: maccode = 0x2F; break; // 5
		case 0x36: maccode = 0x2D; break; // 6
		case 0x3D: maccode = 0x35; break; // 7
		case 0x3E: maccode = 0x39; break; // 8
		case 0x46: maccode = 0x33; break; // 9
		case 0x0E: maccode = 0x65; break; // `
		case 0x4E: maccode = 0x37; break; // -
		case 0x55: maccode = 0x31; break; // =
		case 0x5D: maccode = 0x55; break; // backslash
		case 0x66: maccode = 0x67; break; // bkspc
		case 0x29: maccode = 0x63; break; // space
		case 0x0D: maccode = 0x61; break; // tab
		case 0x58: maccode = 0x73; break; // caps
		case 0x12: maccode = 0x71; break; // left shift
		case 0x59: maccode = 0x71; break; // right shift
		case 0x11: maccode = 0x75; break; // left alt = opt
		case 0x5A: maccode = 0x49; break; // enter
		case 0x41: maccode = 0x57; break; // ,
		case 0x49: maccode = 0x5F; break; // .
		case 0x4A: maccode = 0x59; break; // /
		default: break;
		}
	} else {
		switch(code) {
		case 0x1F: maccode = 0x6F; break; // left windows key = cmd
		case 0x27: maccode = 0x6F; break; // right windows key = cmd
		case 0x11: maccode = 0x75; break; // right alt = cmd
		case 0x5A: maccode = 0x69; break; // keypad enter = enter
		}
	}

	if(keyup) {
		maccode |= 0x80;
	}

	return maccode;
}

void SetPS2LEDState(uint8_t cmd)
{
	unsigned long tmp1 = tenustimer;
	while(tenustimer < (tmp1+10));
	SendPS2Command(0xED);
	// wait for ack 0xFA
	PollPS2Frame();
	tmp1 = tenustimer;
	while(tenustimer < (tmp1+10));
	SendPS2Command(cmd);
	PollPS2Frame();
}

void SendPS2Command(uint8_t cmd)
{
	GPIO_InitTypeDef GPIO_InitStructure;
	unsigned long tmp1;
	uint8_t bits;
	uint8_t parity = 0;

	GPIO_InitStructure.GPIO_Pin = kPS2KbdClockPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	tmp1 = tenustimer;
	GPIO_ResetBits(GPIOC, kPS2KbdClockPin);
	// pull clock low for >=100us
	while(tenustimer < (tmp1 + 15));

	// Bring data line low
	GPIO_InitStructure.GPIO_Pin = kPS2KbdDataPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_OUT;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);
	GPIO_ResetBits(GPIOC, kPS2KbdDataPin);
	
	// Release clock
	GPIO_InitStructure.GPIO_Pin = kPS2KbdClockPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	// Might need to wait for the clock to go high after
	// releasing it...
	while(GPIO_ReadInputDataBit(GPIOC, kPS2KbdClockPin) == 0);

#if 0
	// start bit
	while(GPIO_ReadInputDataBit(GPIOC, kPS2KbdClockPin) == 1);
	while(GPIO_ReadInputDataBit(GPIOC, kPS2KbdClockPin) == 0);
#endif

	for(bits = 0; bits < 8; bits++) {
		// Wait for device to bring clock low
		while(GPIO_ReadInputDataBit(GPIOC, kPS2KbdClockPin) == 1);
		if(cmd & (1<<bits)) {
			GPIO_SetBits(GPIOC, kPS2KbdDataPin);
			parity++;
		} else {
			GPIO_ResetBits(GPIOC, kPS2KbdDataPin);
		}
		// wait out the remainder of the low clock
		while(GPIO_ReadInputDataBit(GPIOC, kPS2KbdClockPin) == 0);
	}

	// Data has been sent, now send parity
	while(GPIO_ReadInputDataBit(GPIOC, kPS2KbdClockPin) == 1);
	if((parity & 0x01) == 0) {
		GPIO_SetBits(GPIOC, kPS2KbdDataPin);
	} else {
		GPIO_ResetBits(GPIOC, kPS2KbdDataPin);
	}
	while(GPIO_ReadInputDataBit(GPIOC, kPS2KbdClockPin) == 0);
	while(GPIO_ReadInputDataBit(GPIOC, kPS2KbdClockPin) == 1);

	// Stop bit, release data line
	GPIO_InitStructure.GPIO_Pin = kPS2KbdDataPin;
	GPIO_InitStructure.GPIO_Mode = GPIO_Mode_IN;
	GPIO_InitStructure.GPIO_OType = GPIO_OType_PP;
	GPIO_InitStructure.GPIO_Speed = GPIO_Speed_50MHz;
	GPIO_InitStructure.GPIO_PuPd = GPIO_PuPd_UP;
	GPIO_Init(GPIOC, &GPIO_InitStructure);

	while(GPIO_ReadInputDataBit(GPIOC, kPS2KbdDataPin) == 0);
	while(GPIO_ReadInputDataBit(GPIOC, kPS2KbdDataPin) == 1);
	while(GPIO_ReadInputDataBit(GPIOC, kPS2KbdDataPin) == 0);
}

uint16_t PollPS2Frame()
{
	uint16_t ps2frame = 0;
	uint8_t bits;

	for(bits = 0; bits < 11; bits++) {
		while(GPIO_ReadInputDataBit(GPIOC, kPS2KbdClockPin) == 1);
		while(GPIO_ReadInputDataBit(GPIOC, kPS2KbdClockPin) == 0);
		// read when clock is high
		ps2frame |= GPIO_ReadInputDataBit(GPIOC, kPS2KbdDataPin) << bits;
	}

	return ps2frame;
}

void HandleMacCommand()
{
	switch(mackbd.cmd) {
	case 0x16:
		mackbd.state = kMacKbdStateWriteSettle;
		mackbd.delay = 4;
		mackbd.response = 0x03;
		break;
	case 0x10:
		mackbd.state = kMacKbdStateInquiryWait;
		mackbd.delay = 25000;
		break;
	default:
		break;
	};

	return;
}

void USART2_SendHexByte(uint8_t c)
{
	const char base16table[16] = { '0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F' };
	while (USART_GetFlagStatus(USART2, USART_FLAG_TXE) == RESET);
	USART_SendData(USART2, base16table[(c >> 4) & 0x0F]);
	while (USART_GetFlagStatus(USART2, USART_FLAG_TXE) == RESET);
	USART_SendData(USART2, base16table[c & 0x0F]);
}

void USART2_SendString(const char *str)
{
	int i;
	for(i = 0; str[i]; i++) {
		while (USART_GetFlagStatus(USART2, USART_FLAG_TXE) == RESET);
		USART_SendData(USART2, str[i]);
	}
}

void USART2_IRQHandler(void)
{
	if (USART_GetITStatus(USART2, USART_IT_RXNE) != RESET) {
		unsigned char c = USART_ReceiveData(USART2);
	}
}
